@@ -55,6 +55,39 @@ def del_all(mapping, to_remove):
     return mapping
 
 
+def get_list_unique_exchanges(databases: list) -> list:
+    """
+    Gets a list of all unique exchanges found in a list of databases.
+
+    :param databases: A list of Brightway2-style databases to extract unique exchanges from.
+    :return: A list of tuples representing unique exchanges, where each tuple contains the following information:
+        - name: The name of the exchange.
+        - product: The reference product of the exchange.
+        - categories: The categories of the exchange.
+        - location: The location of the exchange.
+        - unit: The unit of the exchange.
+        - type: The type of the exchange (either "technosphere" or "biosphere").
+    """
+
+    return list(
+        set(
+            [
+                (
+                    exchange["name"],
+                    exchange.get("product"),
+                    exchange.get("categories"),
+                    exchange.get("location"),
+                    exchange.get("unit"),
+                    exchange.get("type"),
+                )
+                for database in databases
+                for dataset in database
+                for exchange in dataset["exchanges"]
+            ]
+        )
+    )
+
+
 class Unfold:
     """Extracts datapackage files."""
 
@@ -73,6 +106,7 @@ class Unfold:
         self.databases_to_export = {}
         self.dependency_mapping = {}
         self.factors = {}
+        self.name = None
 
     def show_scenarios(self):
         """Shows the scenarios."""
@@ -266,7 +300,10 @@ class Unfold:
         """Extracts additional inventories."""
         print("Extracting additional inventories...")
         with HiddenPrints():
-            i = bw2io.CSVImporter(self.package.get_resource("inventories").source)
+            filepath = self.package.get_resource("inventories").source
+            # ensure that all slashes in filepath are forward slashes
+            filepath = filepath.replace("\\", "/")
+            i = bw2io.CSVImporter(filepath)
             i.strategies = i.strategies[:4]
             i.apply_strategies()
             i.data = self.clean_imported_inventory(i.data)
@@ -290,38 +327,6 @@ class Unfold:
             self.scenario_df.groupby("flow id").sum(numeric_only=True).to_dict("index")
         )
 
-    def get_list_unique_exchanges(self, databases: list) -> list:
-        """
-        Gets a list of all unique exchanges found in a list of databases.
-
-        :param databases: A list of Brightway2-style databases to extract unique exchanges from.
-        :return: A list of tuples representing unique exchanges, where each tuple contains the following information:
-            - name: The name of the exchange.
-            - product: The reference product of the exchange.
-            - categories: The categories of the exchange.
-            - location: The location of the exchange.
-            - unit: The unit of the exchange.
-            - type: The type of the exchange (either "technosphere" or "biosphere").
-        """
-
-        return list(
-            set(
-                [
-                    (
-                        exchange["name"],
-                        exchange.get("product"),
-                        exchange.get("categories"),
-                        exchange.get("location"),
-                        exchange.get("unit"),
-                        exchange.get("type"),
-                    )
-                    for database in databases
-                    for dataset in database
-                    for exchange in dataset["exchanges"]
-                ]
-            )
-        )
-
     def store_datasets_metadata(self) -> None:
         """
         Stores metadata for each dataset in the database attribute.
@@ -359,7 +364,7 @@ class Unfold:
     def generate_activities_indices(self) -> None:
         """Generates the indices of the unique activities."""
         # Get all unique activities from self.database.
-        list_unique_acts = self.get_list_unique_exchanges(databases=[self.database])
+        list_unique_acts = get_list_unique_exchanges(databases=[self.database])
 
         # Add additional exchanges to list_unique_acts.
         for act in list_unique_acts:
@@ -388,6 +393,55 @@ class Unfold:
         else:
             return str(uuid.uuid4().hex)
 
+    def find_correct_id(self, key: tuple) -> [tuple, None]:
+        if key[0] in self.outdated_flows:
+            if (
+                self.outdated_flows[key[0]],
+                key[1],
+                key[2],
+                key[3],
+            ) in self.dependency_mapping:
+                return self.outdated_flows[key[0]], key[1], key[2], key[3]
+
+            elif (
+                self.outdated_flows[key[0]].replace(", ion", ""),
+                key[1],
+                key[2],
+                key[3],
+            ) in self.dependency_mapping:
+                return (
+                    self.outdated_flows[key[0]].replace(", ion", ""),
+                    key[1],
+                    key[2],
+                    key[3],
+                )
+
+            elif (
+                self.outdated_flows[key[0]] + ", ion",
+                key[1],
+                key[2],
+                key[3],
+            ) in self.dependency_mapping:
+                return self.outdated_flows[key[0]] + ", ion", key[1], key[2], key[3]
+
+            elif (
+                key[0] + ", in ground",
+                key[1],
+                key[2],
+                key[3],
+            ) in self.dependency_mapping:
+                return key[0] + ", in ground", key[1], key[2], key[3]
+
+            else:
+                print("Could not find key", key)
+                return key
+
+    def fix_key(self, key: tuple) -> tuple:
+        if key in self.dependency_mapping:
+            return self.dependency_mapping[key]
+        else:
+            return self.dependency_mapping[self.find_correct_id(key)]
+
     def get_exchange(
         self,
         ind: int,
@@ -407,14 +461,6 @@ class Unfold:
         name, ref, cat, loc, unit, flow_type = self.acts_indices[ind]
         _ = lambda x: x if x != 0 else 1.0
 
-        def fix_key(key: tuple) -> tuple:
-            if key in self.dependency_mapping:
-                return self.dependency_mapping[key]
-            else:
-                return self.dependency_mapping[
-                    (self.outdated_flows.get(key[0], key[0]), key[1], key[2], key[3])
-                ]
-
         return {
             "name": name,
             "product": ref,
@@ -423,7 +469,7 @@ class Unfold:
             "categories": cat,
             "type": flow_type,
             "amount": amount if flow_type != "production" else _(amount),
-            "input": fix_key((name, ref, loc, cat))
+            "input": self.fix_key((name, ref, loc, cat))
             if flow_type == "biosphere"
             else (
                 scenario_name,
@@ -511,7 +557,7 @@ class Unfold:
             s_name, s_prod, s_loc, s_cat, s_unit, s_type = list(flow_id)[4:]
 
             # Look up the index of the consumer activity in the reversed activities index.
-            consumer_idx = self.reversed_acts_indices[
+            consumer_idx = self.reversed_acts_indices.get(
                 (
                     c_name,
                     c_prod,
@@ -519,8 +565,9 @@ class Unfold:
                     c_loc,
                     c_unit,
                     "production",
-                )
-            ]
+                ),
+                None,
+            )
 
             # Look up the index of the supplier activity in the reversed activities index.
             supplier_id = (
@@ -531,13 +578,21 @@ class Unfold:
                 s_unit,
                 s_type,
             )
-            supplier_idx = self.reversed_acts_indices[supplier_id]
+            supplier_idx = self.reversed_acts_indices.get(
+                supplier_id,
+                None,
+            )
 
             # Multiply the appropriate element of the matrix by the scaling factor for the given scenario.
             # Use the lambda function defined above to avoid multiplying by zero.
-            matrix[supplier_idx, consumer_idx] = factor[scenario_name] * _(
-                matrix[supplier_idx, consumer_idx]
-            )
+            if supplier_idx is not None and consumer_idx is not None:
+                matrix[supplier_idx, consumer_idx] = factor[scenario_name] * _(
+                    matrix[supplier_idx, consumer_idx]
+                )
+            else:
+                print(
+                    f"Could not find activity for flow {flow_id} in scenario {scenario_name}."
+                )
 
         # Return the scaled matrix.
         return matrix
@@ -602,7 +657,7 @@ class Unfold:
             # Get the dictionary representing the current activity from the database.
             act = self.get_act_dict_structure(
                 ind=k,
-                scenario_name=self.package.descriptor["name"],
+                scenario_name=self.name or self.package.descriptor["name"],
             )
 
             # Update the activity dictionary with metadata.
@@ -613,7 +668,7 @@ class Unfold:
                 self.get_exchange(
                     ind=j,
                     amount=matrix[j, k, 0],
-                    scenario_name=self.package.descriptor["name"],
+                    scenario_name=self.name or self.package.descriptor["name"],
                 )
                 for j in v
             )
@@ -804,9 +859,10 @@ class Unfold:
         )
 
         # Load the scenario dataframe from the `scenario_data` resource.
-        self.scenario_df = pd.DataFrame(
-            self.package.get_resource("scenario_data").read(keyed=True)
-        )
+        filepath = self.package.get_resource("scenario_data").source
+        # ensure that all slashes in filepath are forward slashes
+        filepath = filepath.replace("\\", "/")
+        self.scenario_df = pd.read_csv(filepath, keep_default_na=False, na_values="")
 
         # Remove rows corresponding to production flows.
         self.scenario_df = self.scenario_df.loc[
@@ -827,9 +883,11 @@ class Unfold:
         self.scenario_df["to categories"] = self.scenario_df["to categories"].apply(
             lambda x: literal_eval(str(x))
         )
+
         self.scenario_df["from key"] = self.scenario_df["from key"].apply(
             lambda x: literal_eval(str(x))
         )
+
         self.scenario_df["to key"] = self.scenario_df["to key"].apply(
             lambda x: literal_eval(str(x))
         )
@@ -886,6 +944,7 @@ class Unfold:
                     "production",
                 )
             ]
+
             supplier_id = (
                 s_name,
                 s_prod,
@@ -894,6 +953,27 @@ class Unfold:
                 s_unit,
                 s_type,
             )
+
+            if supplier_id not in self.reversed_acts_indices:
+                try:
+                    (
+                        s_name,
+                        s_prod,
+                        s_loc,
+                        s_cat,
+                    ) = self.find_correct_id((s_name, s_prod, s_loc, s_cat))
+
+                    supplier_id = (
+                        s_name,
+                        s_prod,
+                        s_cat,
+                        s_loc,
+                        s_unit,
+                        s_type,
+                    )
+                except TypeError:
+                    print("not found key for", supplier_id)
+
             supplier_idx = self.reversed_acts_indices[supplier_id]
 
             # Update the factor values for each scenario by multiplying with the corresponding matrix value
@@ -903,7 +983,9 @@ class Unfold:
         # Create a new scenario dataframe from the updated factors dictionary
         self.scenario_df = pd.DataFrame.from_dict(self.factors).T.reset_index()
 
-        # Rename columns and add new columns for database information and metadata
+        # Rename columns
+        # and add new columns
+        # for database information and metadata
         self.scenario_df.columns = [
             "to activity name",
             "to reference product",
@@ -917,13 +999,22 @@ class Unfold:
             "flow type",
         ] + [s["name"] for s in self.scenarios]
 
-        self.scenario_df["to database"] = self.package.descriptor["name"]
+        self.scenario_df["to database"] = self.name or self.package.descriptor["name"]
         self.scenario_df["to categories"] = None
         self.scenario_df["to key"] = None
         self.scenario_df["from key"] = None
         self.scenario_df = self.scenario_df.replace({np.nan: None})
 
-        # Use the dependency_mapping dictionary to add "from key" and "to key" columns
+        # if self.name is not None, we change the name of the database
+        # in self.dependency_mapping to the name of the superstructure
+        if self.name:
+            self.dependency_mapping = {
+                k: (self.name, v[1]) if v[0] == self.package.descriptor["name"] else v
+                for k, v in self.dependency_mapping.items()
+            }
+
+        # Use the dependency_mapping dictionary
+        # to add "from key" and "to key" columns
         self.scenario_df.loc[:, "from key"] = self.scenario_df.apply(
             lambda x: self.dependency_mapping.get(
                 (
@@ -950,7 +1041,7 @@ class Unfold:
         # Add "from database" column based on flow type
         self.scenario_df.loc[
             (self.scenario_df["flow type"] == "technosphere"), "from database"
-        ] = self.package.descriptor["name"]
+        ] = (self.name or self.package.descriptor["name"])
         self.scenario_df.loc[
             (self.scenario_df["flow type"] == "biosphere"), "from database"
         ] = "biosphere3"
@@ -978,6 +1069,7 @@ class Unfold:
         scenarios: List[int] = None,
         dependencies: dict = None,
         superstructure: bool = False,
+        name: str = None,
     ):
         """
         unfold() is a method of the Unfold class, which extracts specific scenarios from the input LCA database and writes them as new databases.
@@ -985,6 +1077,7 @@ class Unfold:
         :param scenarios: A list of integers indicating the indices of the scenarios to extract. If None, all scenarios are extracted. Default is None.
         :param dependencies: A dictionary containing additional inventory databases that may be needed for extraction. Default is None.
         :param superstructure: A boolean indicating whether to generate a scenario difference file and a superstructure database. Default is False.
+        :param name: A string indicating the name of the superstructure database. Default is None.
         :return: None
 
         Behavior:
@@ -996,10 +1089,11 @@ class Unfold:
         Calls the generate_factors() method to generate the factors for the scenarios.
         If superstructure is False, calls the generate_single_databases() method to generate a separate LCA database for each scenario, and stores them in the databases_to_export dictionary.
         If superstructure is True, calls the format_superstructure_dataframe() method to generate a scenario difference file and a superstructure database.
-        Calls the write() method to write the databases to disk.
+        Calls write() method to write the databases to disk.
 
         """
 
+        self.name = name
         self.check_dependencies(dependencies)
         self.extract_source_database()
         self.extract_additional_inventories()
@@ -1046,7 +1140,7 @@ class Unfold:
         else:
             try:
                 self.scenario_df.to_excel(
-                    f"{self.package.descriptor['name']}.xlsx", index=False
+                    f"{self.name or self.package.descriptor['name']}.xlsx", index=False
                 )
             except ValueError:
                 # from https://stackoverflow.com/questions/66356152/splitting-a-dataframe-into-multiple-sheets
@@ -1064,14 +1158,16 @@ class Unfold:
             )
             print("")
             print("Writing superstructure database...")
-            change_db_name(self.database, self.package.descriptor["name"])
+            change_db_name(self.database, self.name or self.package.descriptor["name"])
             self.database = check_exchanges_input(
                 self.database, self.dependency_mapping
             )
             link_internal(self.database)
             check_internal_linking(self.database)
             check_duplicate_codes(self.database)
-            correct_fields_format(self.database, self.package.descriptor["name"])
+            correct_fields_format(
+                self.database, self.name or self.package.descriptor["name"]
+            )
             UnfoldExporter(
-                self.package.descriptor["name"], self.database
+                self.name or self.package.descriptor["name"], self.database
             ).write_database()
